# %%
import os
import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import datasets, transforms


class ConvBlock(nn.Module):
    def __init__(self, in_channels : int, out_channels : int) -> None:
        super().__init__()

        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channels, out_channels, kernel_size=3, padding='same'),
            nn.GroupNorm(1, out_channels),   
            nn.GELU()
        )

        self.conv2 = nn.Sequential(
            nn.Conv2d(out_channels, out_channels, kernel_size=3, padding='same'),
            nn.GroupNorm(1, out_channels),   
            nn.GELU()
        )
    
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)

        return x

# %%
## Making a Unet model using the Basic Convolution Block shown above

class Unet(nn.Module):
    def __init__(self, in_channels : int, n_filters = 32, n_class = 5, add_sigmoid = False):
        super().__init__()

        self.add_sigmoid = add_sigmoid

        ## Down Sampling
        self.dconv1 = ConvBlock(in_channels, n_filters)
        self.maxpool1 = nn.MaxPool2d(kernel_size= 2, stride= 2)
        self.dconv2 = ConvBlock(n_filters, n_filters * 2)
        self.maxpool2 = nn.MaxPool2d(kernel_size= 2, stride= 2)
        #self.dconv3 = ConvBlock(n_filters * 2, n_filters * 4)
        #self.maxpool3 = nn.MaxPool2d(kernel_size= 2, stride= 2)
        #self.dconv4 = ConvBlock(n_filters * 4, n_filters * 8)
        #self.dropout4 = nn.Dropout(p = 0.3)
        #self.maxpool4 = nn.MaxPool2d(kernel_size= 2, stride= 2)

        # I just want 2 layers for now
        

        ## Bottleneck Layer
        self.bottleneck = ConvBlock(n_filters * 2, n_filters * 4)
        self.dropout_bottle = nn.Dropout(p = 0.3)


        ### Up sampling
        #self.trans_conv4 = nn.ConvTranspose2d(n_filters * 16, n_filters * 8, kernel_size= 2,
        #                                   stride= 2)
        ## concat step between trans_conv4, and uconv4, but will be shown in forward fn.
        #self.uconv4 = ConvBlock(n_filters * 16, n_filters * 8)

        #self.trans_conv3 = nn.ConvTranspose2d(n_filters * 8, n_filters * 4, kernel_size= 2,
        #                                   stride= 2)
        ## concat step between trans_conv4, and uconv4, but will be shown in forward fn.
        #self.uconv3 = ConvBlock(n_filters * 8, n_filters * 4)

        self.trans_conv2 = nn.ConvTranspose2d(n_filters * 4, n_filters * 2, kernel_size= 2,
                                           stride= 2)
        # concat step between trans_conv4, and uconv4, but will be shown in forward fn.
        self.uconv2 = ConvBlock(n_filters * 4, n_filters * 2)

        self.trans_conv1 = nn.ConvTranspose2d(n_filters * 2, n_filters * 1, kernel_size= 2,
                                           stride= 2)
        # concat step between trans_conv4, and uconv4, but will be shown in forward fn.
        self.uconv1 = ConvBlock(n_filters * 2, n_filters * 1)

        ## Final class allocations
        self.orig_dims = nn.Conv2d(in_channels= n_filters, out_channels= in_channels, kernel_size= 3,
                                    padding= 'same')
        self.out_class_dims = nn.Conv2d(in_channels= in_channels, out_channels= n_class, 
                                         kernel_size= 1, padding= 'same')
        
        if self.add_sigmoid:
            self.out_sigmoid = nn.Sigmoid() # passing sigmoid function to bring output between 0 and 1
        

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        
        ## Downsampling
        x = self.dconv1(x)
        skip1 = x.clone()
        x = self.maxpool1(x)
        x = self.dconv2(x)
        skip2 = x.clone()
        x = self.maxpool2(x)
        #x = self.dconv3(x)
        #skip3 = x.clone()
        #x = self.maxpool3(x)
        #x = self.dconv4(x)
        ##x = self.dropout4(x)
        #skip4 = x.clone()
        #x = self.maxpool4(x)


        ## Bottleneck layer
        x = self.bottleneck(x)
        #x = self.dropout_bottle(x)


        ## Upsampling
        #x = self.trans_conv4(x)
        #x = torch.cat([x, skip4], dim = 1)
        #x = self.uconv4(x)


        #x = self.trans_conv3(x)
        #x = torch.cat([x, skip3], dim = 1)
        #x = self.uconv3(x)


        x = self.trans_conv2(x)
        x = torch.cat([x, skip2], dim = 1)
        x = self.uconv2(x)


        x = self.trans_conv1(x)
        x = torch.cat([x, skip1], dim = 1)
        x = self.uconv1(x)


        ## Final class allocations
        x = self.orig_dims(x)
        x = self.out_class_dims(x)

        ## Adding sigmoid layer at end if it needs to be added
        if self.add_sigmoid:
            x = self.out_sigmoid(x)

        
        return x


def test(channels, height, width):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Using device: {device}")

    x = torch.randn((10, channels, height, width)).to(device)      # [# examples, channels, height, width]
    model = Unet(in_channels= x.shape[1], n_class= 4, n_filters= 32).to(device)
    preds = model(x)
    print(f"output shape: {preds.shape}")
    print(f"input shape: {x.shape}")

if __name__ == '__main__':
    
    channels = input("Enter # of channels: ")
    height = input("Enter image height: ")
    width = input("Enter image width: ")

    test(int(channels), int(height), int(width))
# %%
