import os
import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
import matplotlib.pyplot as plt

import torch
import torchvision.transforms as transforms
import numpy as np
from torch.utils.data import Dataset, DataLoader
import os

class CustomDataset(Dataset):
    def __init__(self, root_dir):
        self.root_dir = root_dir
        self.file_list = self._prepare_file_list()
    
    def normalize_tensor(self, tensor):
        mean = torch.mean(tensor)
        std = torch.std(tensor)

        # Check if std is zero or very close to zero
        if std.item() == 0 or std.item() < 1e-9:
            # print("Warning: Standard deviation is zero or too small")
            # print(f"The mean in this scenario is: {mean}")
            normalized = tensor  # Decide how to handle this case
        else:
            normalized = (tensor - mean) / std

        return normalized, mean, std


    def _prepare_file_list(self):
        mixture_dir = os.path.join(self.root_dir, 'Mixture')
        mix_files = sorted(os.listdir(mixture_dir))
        file_list = []
        for mix_file in mix_files:
            file_id = os.path.splitext(mix_file)[0].split('_')[0]
            file_paths = {'Mixture': os.path.join(mixture_dir, mix_file)}
            for stream in ['Accompaniment', 'Bass', 'Drums', 'Vocals']:
                stream_file = f'{file_id}_{stream}.npy'
                file_paths[stream] = os.path.join(self.root_dir, stream, stream_file)
            file_list.append(file_paths)
        #print(file_list)
        return file_list

    def __len__(self):
        return len(self.file_list)

    def __getitem__(self, idx):
        file_paths = self.file_list[idx]
        norm_params = {}  # initializing dict to store mean and std for each stream
        accompaniment = None
        bass = None
        drums = None
        vocals = None

        # iterating through each stream's numpy file
        for stream, path in file_paths.items():
            #print(f"Stream: {stream} and Path: {path}")
            arr = np.load(path)
            arr_tensor = torch.from_numpy(arr[np.newaxis, :, :]).float()  # Convert to tensor here
            #print(f"Does the numpy array have nan values: {np.isnan(arr).any()}")
            #print(f"Does the tensor array have any nan values: {torch.isnan(arr_tensor).any()}")

            normalized_tensor, mean, std = self.normalize_tensor(arr_tensor)  # Normalize data
            #print(f"Does the normalized tensor array have any nan values: {torch.isnan(normalized_tensor).any()}")

            # Store mean and std for each file's stream in norm_params
            norm_params[stream] = {'mean': mean.item(), 'std': std.item()}  # Convert mean and std to Python scalars
            #print(F"Normalization Params for {stream}: {norm_params[stream]}")

            if stream == 'Mixture':
                mixture = arr_tensor  # Training data (Mixture)
            elif stream == 'Accompaniment':
                accompaniment = arr_tensor
            elif stream == 'Bass':
                bass = arr_tensor
            elif stream == 'Drums':
                drums = arr_tensor
            elif stream == 'Vocals':
                vocals = arr_tensor

        # Convert list of tensors to a single tensor for the target values
        indv_streams = torch.cat([accompaniment, bass, drums, vocals], dim=0)

        return mixture, indv_streams, norm_params


def de_normalize(tensor, mean, std):
    """Reverses normalization for a tensor using given mean and std."""
    return tensor * std + mean

def find_nan(arr_tensor):

    # Flatten the tensor
    tensor_reshaped = arr_tensor.reshape(arr_tensor.shape[0], -1)

    # Create a boolean mask for NaN values
    nan_mask = torch.isnan(tensor_reshaped)

    # Reshape the boolean mask back to the original shape
    nan_mask = nan_mask.reshape(nan_mask.shape[0], *arr_tensor.shape[1:])

    # Print whether there are NaN values in the tensor
    # if torch.any(nan_mask):
    #     print("The tensor contains NaN values.")
        
    #     # Print the boolean mask (True where NaN, False otherwise)
    #     print("Boolean mask for NaN values:")
    #     print(nan_mask)
    # else:
    #     print("The tensor does not contain any NaN values.")

def check_data_integrity(dataloader):
    x_shape, y_shape = None, None
    for batch, (X, y, norm_params) in enumerate(dataloader):
        # Check for NaN values in X (assuming X can have multiple streams like Y, if not, adjust accordingly)
        for i in range(X.size(1)): # Assuming X is shaped [batch, channels/streams, height, width]
            if torch.isnan(X[:, i, :, :]).any():
                print(f"Found NaN values in X stream {i} of batch {batch}")
                return False

        # Check for NaN values in Y (assuming Y consists of multiple streams)
        for stream_idx, stream_tensor in enumerate(y): # y shaped [num_streams, batch, height, width]
            if torch.isnan(stream_tensor).any():
                print(f"Found NaN values in Y stream {stream_idx} of batch {batch}")
                return False

        # Check consistent shape
        if x_shape is None and y_shape is None:
            x_shape, y_shape = X.shape[1:], y.shape[1:]  # Update to ignore batch dimension
        elif x_shape != X.shape[1:] or y_shape != y.shape[1:]:
            print(f"Inconsistent shape found in batch {batch}, X shape: {X.shape}, Y shape: {y.shape}")
            return False

    print("Data integrity check passed")
    return True


if __name__ == "__main__":

    # Usage
    root_dir = r'data/trimini_processed'
    dataset = CustomDataset(root_dir)
    dataloader = DataLoader(dataset, batch_size=2, shuffle=True)

    next(iter(dataloader))

    #check_data_integrity(dataloader)

    # Iterating through the DataLoader (example)
    for batch, (X, y, norm_params) in enumerate(dataloader):
        # Each 'batch' is a dictionary with keys 'Mixture', 'Accompaniment', 'Bass', 'Drums', 'Vocals'
        # The value for each key is a tensor of shape [batch_size, *, *, ...] depending on your data's shape
        mixture = X
        indv_streams = y
        norm_params = norm_params
        # Process the batch...
        break  # Just to break the loop, remove in actual usage

    print(f"Mixture Shape: {mixture.shape}")
    #print(f"Mixture: {mixture}")

    print(f"Streams Shape: {indv_streams.shape}")
    #print(f"Streams: {indv_streams}")

    print(f"Normalization Parameters Shape: {norm_params.keys()}")
    print(f"Normalization Parameters: {norm_params}")
    #print(f"Normalization Parameter Access trial: {norm_params['Mixture']['mean'][0]}")

    print 

    
    tensor_normalized = mixture[0][0].detach().cpu().numpy()  # Convert to numpy array
    print(f"tensor_normalized max and min: {tensor_normalized.max()} and {tensor_normalized.min()}")
    #print(f"tensor_normalized: {tensor_normalized}")


    # plt.figure(figsize=(10, 5))
    # plt.imshow((tensor_normalized + 1e-9), aspect='auto', origin='lower', cmap='jet')
    # plt.colorbar()
    # plt.title('STFT Magnitude Normalized')
    # plt.xlabel('Frames')
    # plt.ylabel('Frequency Bins')
    # plt.tight_layout()
    # plt.show()

    # tensor_denormalized = de_normalize(mixture[0][0], norm_params['Mixture']['mean'][0], norm_params['Mixture']['std'][0])
    # tensor_denormalized = tensor_denormalized.detach().cpu().numpy()  # Convert to numpy array

    # plt.figure(figsize=(10, 5))
    # plt.imshow((tensor_normalized + 1e-9), aspect='auto', origin='lower', cmap='jet')
    # plt.colorbar()
    # plt.title('STFT Magnitude De-Normalized')
    # plt.xlabel('Frames')
    # plt.ylabel('Frequency Bins')
    # plt.tight_layout()
    # plt.show()
