import os
import librosa
import numpy as np
import stempeg
import matplotlib.pyplot as plt

def process_and_save_mel(input_dir, output_dir, n_fft=512, hop_length=512, n_mels=128, target_size=(128, 256)):
    # Define stream names
    stream_names = ['Mixture', 'Drums', 'Bass', 'Accompaniment', 'Vocals']

    # Ensure output subfolders exist
    for name in stream_names:
        os.makedirs(os.path.join(output_dir, name), exist_ok=True)
    
    # Initialize a counter for naming files numerically and plotting
    file_counter = 0
    plot_counter = 0  # Counter for plotting
    
    # Process each file in the input directory
    for file_name in os.listdir(input_dir):
        if plot_counter >= 20:  # Exit the loop after plotting 20 images
            break
        
        file_path = os.path.join(input_dir, file_name)
 
        # Read all streams from the stem file
        streams, sr = stempeg.read_stems(file_path, always_3d=True)
        
        for i, name in enumerate(stream_names):
            if plot_counter >= 20:  # Check again inside the inner loop
                break
            
            # Convert to mono
            mono_audio = np.mean(streams[i], axis=-1)
            
            # Compute Mel spectrogram
            mel_spec = librosa.feature.melspectrogram(y=mono_audio, sr=sr, n_fft=n_fft, hop_length=hop_length, n_mels=n_mels)
            # Convert to dB
            mel_spec_db = librosa.power_to_db(mel_spec, ref=np.max)
            
            # Trim the Mel spectrogram array to target size
            trimmed_mel = mel_spec_db[:target_size[0], :target_size[1]]

            if trimmed_mel.shape == target_size and not np.isnan(trimmed_mel).any():
                # Define save path with numerical naming
                save_path = os.path.join(output_dir, name, f'{file_counter}_{name}.npy')
                # Save Mel spectrogram
                np.save(save_path, trimmed_mel)
                print(f"Processed and saved Mel spectrograms for file number {file_counter}, stream: {name}")
                
                # Plotting
                plt.figure(figsize=(10, 4))
                plt.imshow(trimmed_mel, aspect='auto', origin='lower', cmap='viridis')
                plt.title(f"Mel Spectrogram for {name} (file {file_counter})")
                plt.colorbar(format='%+2.0f dB')
                plt.show()
                
                plot_counter += 1
                
            elif trimmed_mel.shape != target_size:
                print(f"Skipped file number {file_counter}, stream: {name} due to mismatched size.")
            else:
                print(f"Found NaN values in Mel spectrogram for file number {file_counter}, stream: {name}")
        
        file_counter += 1

# Example usage
input_dir = "C:/Users/andre/EE367/audio_separation_unet_drew/data/trimmed_audio"
output_dir = "C:/Users/andre/EE367/audio_separation_unet_drew/data/trimmed_audio_Mel_test"
process_and_save_mel(input_dir, output_dir, n_fft=512, hop_length=512, n_mels=128, target_size=(128, 256))
