import librosa
import numpy as np
import soundfile as sf
import plotly.graph_objects as go
import stempeg
import os
from plotly.subplots import make_subplots
import matplotlib.pyplot as plt

###############################################################################################################################

def plot_stft_subplots(audio_file_path):
    """
    Plots the STFT magnitude (in dB) for different audio streams as subplots using matplotlib.
    
    Parameters:
    - audio_file_path: Path to the directory containing the audio files.
    """
    track_titles = ['Mixture', 'Drums', 'Bass', 'Accompaniment', 'Vocals']
    tracks = [os.path.join(audio_file_path, f"{title}.flac") for title in track_titles]
    
    # Create subplots
    fig, axes = plt.subplots(nrows=len(tracks), ncols=1, figsize=(15, 10), sharex=True)
    
    for i, (track_path, title) in enumerate(zip(tracks, track_titles)):
        # Load an audio file
        y, sr = librosa.load(track_path, sr=None, mono=True)

        # Compute the STFT of the audio
        stft = librosa.stft(y, n_fft=2048, hop_length=512)

        # Convert the complex-valued STFT to magnitude for visualization
        D_magnitude = np.abs(stft)

        # Convert magnitude to decibels
        D_db = librosa.amplitude_to_db(D_magnitude, ref=np.max)

        # Plot
        img = librosa.display.specshow(D_db, sr=sr, hop_length=512, x_axis='time', y_axis='log', ax=axes[i])
        axes[i].set_title(title)
        fig.colorbar(img, ax=axes[i], format="%+2.f dB")
    
    plt.tight_layout()
    plt.show()

# Example usage
plot_stft_subplots("split_audio_files")

