import torch
from tqdm import tqdm
import os
from torch.utils.tensorboard import SummaryWriter
from datetime import datetime
import matplotlib.pyplot as plt

eval_print_duration = 1   # After how many batches do you want to print the loss and accuracy to terminal and tensorboard
# Functions for training the model
def train_loop(dataloader, model, loss_fn, optimizer, epochs, do_threshold = False, val_data = None, save_dir = None, device = 'cpu'):

    # Set the model to training mode - important for batch normalization and dropout layers
    size = len(dataloader.dataset)      # Find size of dataset
    model.train()                       # Set model to training mode

    # Setup TensorBoard
    curr_dt = datetime.now()
    curr_dt = curr_dt.strftime("%d-%m-%Y %H-%M")
    writer = SummaryWriter(log_dir=f'runs/source_separation/{curr_dt}') # creates folder based on current date time

    # Not necessary for tensorboard, but nice to be able to view the model
    mixture, other_streams, normalization = next(iter(dataloader))      # Loading Image and Masks to view in TensorBoard
    writer.add_graph(model, mixture.to(device)) # check if images.to(device) works if this doesn't work at all
    writer.close()

    counter = 1
    for ep in range(epochs):
        print(f'epoch {ep}')
        # iterate over the mini batches from the dataloader
        pbar = tqdm(dataloader, mininterval=2)
        for batch, (X, y, norm_params) in enumerate(pbar):

            # Required if training on GPU, both Data and Model needs
            # to be on the same device
            X = X.to(device)
            y = y.to(device)
            norm_params = norm_params

            pred = model(X)             # run inference
            if do_threshold:
                pred = thresholding(pred, 0.5)
            loss = loss_fn(pred, y)     # calculate loss

            # Backprop
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            if counter % eval_print_duration == 0:     # Prints metrics every eval_print_duration steps
                loss, current = loss.item(), (batch + 1) * len(X)
                # print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")
                writer.add_scalar("Training Loss:", loss, global_step = counter)

                accuracy = check_accuracy(y, pred)
                # print(f"Accuracy: {accuracy:>7f}")
                writer.add_scalar("Training Accuracy:", accuracy, global_step = counter)

            counter += 1    # increment counter

        # Perform Validation Check on Validation Dataset
        if val_data is not None:
            val_loop(val_data, model, loss_fn, writer, ep, do_threshold, device)
        
        # save model periodically
        if save_dir is not None:
            if ep%512==0 or ep == int(epochs-1):
                if not os.path.exists(save_dir):
                    os.mkdir(save_dir)
                torch.save(model.state_dict(), os.path.join(save_dir, f"model_{ep}.pth"))
                print('saved model at ' + os.path.join(save_dir, f"model_{ep}.pth"))

# Function for checking val data (not same as running inference)
def val_loop(dataloader, model, loss_fn, writer, epoch, do_threshold, device):

    model.eval()
    size = len(dataloader.dataset)
    num_batches = len(dataloader)
    val_loss, accuracy = 0, 0

    with torch.no_grad():
        pbar = tqdm(dataloader, mininterval=2)
        for X, y, norm_params in pbar:

            X = X.to(device)
            y = y.to(device)
            norm_params = norm_params

            # Predictions, loss and accuracy for each batch
            pred = model(X)             # run inferenceß
            if do_threshold:
                pred = thresholding(pred, 0.5)
            val_loss += loss_fn(pred, y).item()    # loss is summed for all batches
            accuracy += check_accuracy(y, pred)          # accuracy is summed for all batches
            
    val_loss /= num_batches        # Average Loss
    accuracy /= num_batches              # Average Accuracy

    # Print to Tensorboard
    writer.add_scalar("Validation Loss:", val_loss, global_step = epoch)
    writer.add_scalar("Validation Accuracy:", accuracy, global_step = epoch)


    print(f"Test Error: \nAverage Accuracy (IOU): {(accuracy):>0.1f}%, Avg loss: {val_loss:>8f} \n")


def check_accuracy(true_mask, pred_mask):
    # Flatten tensors
    true_mask = true_mask.view(-1)
    pred_mask = pred_mask.view(-1)

    # Find intersection by multiplying true_mask and pred_mask (AND)
    intersection = (true_mask * pred_mask).sum()

    # Find union by adding up all true_mask with pred_mask (OR)
    union = true_mask.sum() + pred_mask.sum() - intersection

    iou = 100 * (intersection) / (union + 1e-8) # added a small value to prevent division by zero

    return iou.item()


def dice_coeff(true_mask, pred_mask, is_loss = False):
    smooth = 1.  # Used for 

    # Flatten Tensors
    true_mask = true_mask.view(-1)
    pred_mask = pred_mask.view(-1)

    intersection = (true_mask * pred_mask).sum()

    union = true_mask.sum() + pred_mask.sum()

    dice = (2 * intersection + smooth) / (union + smooth)

    if is_loss:
        return 1 - dice     # loss is 1 - value tensor
    else:
        return dice.item() * 100


def thresholding(mask, threshold):
    return mask * (mask >= threshold)



if __name__ == '__main__':

    print("Unit Test for IOU")

    true_mask = torch.tensor([[1, 1], [0, 0]])
    pred_mask = torch.tensor([[1, 0], [0, 1]])
    calculated_iou = check_accuracy(true_mask, pred_mask)
    expected_iou = 1/3
    
    print(f"Expected IOU: {expected_iou} \nCalculated IOU: {calculated_iou}")


    print('============================')
    print('============================')
    calculated_dice = dice_coeff(true_mask, pred_mask)
    calculated_dice_loss = dice_coeff(true_mask, pred_mask, True)
    
    print(f'Calculated Dice: {calculated_dice}')
    print(f'Calculated Dice Loss: {calculated_dice_loss}')
    print(f"type of loss var: {type(calculated_dice_loss)}")
    print(f"loss var.item(): {calculated_dice_loss.item()}")
